#ifndef PLOG_LOG_MESSAGE_STREAM_H
#define PLOG_LOG_MESSAGE_STREAM_H

#include <iostream>

namespace plog
{
    class msg_writer
    {
    public:
        virtual ~msg_writer(){}
        virtual std::ostream& operator() (std::ostream& os) const = 0;
    };

    inline void hex_write_byte(std::ostream& os, unsigned char b)
    {
        const char* hexbytes[] = {"0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "A", "B", "C", "D", "E", "F"};
        os << (hexbytes[b >> 4]) << (hexbytes[b & 0x0F]);
    }

    template <typename T>
    class hex_writer: public msg_writer
    {
    public:
        hex_writer(T value)
            : value_(value)
        {
        }

        std::ostream& operator() (std::ostream& os) const
        {
            const unsigned char* buffer = reinterpret_cast<const unsigned char*>(&value_);
            for (size_t i = sizeof(T); i != 0; --i)
            {
                hex_write_byte(os, buffer[i-1]);
            }
            return os;
        }

    private:
        T value_;
    };

    class message_buffer: public std::streambuf
    {
        const static size_t BUF_SIZE = 512;
    public:
        message_buffer()
            : position_(0)
        {
            setp(buffer_, buffer_ + BUF_SIZE);
        }

        const char* first() const
        {
            return pbase();
        }

        const char* last() const
        {
            return pptr();
        }

    private:
        char buffer_[BUF_SIZE];
        size_t position_;
    };

    class message_stream: public std::ostream
    {
        friend inline message_stream& operator<< (message_stream& out, const wchar_t* chunk);
        friend message_stream& operator<< (message_stream& out, const std::wstring& chunk);

    public:
        message_stream(const std::locale& loc, const char* s = "")
            : std::ostream(&buf_)
            , locale_(loc)
        {
            *this << s;
        }

        std::string str() const
        {
            return std::string(first(), last());
        }

        const char* first() const
        {
            return buf_.first();
        }

        const char* last() const
        {
            return buf_.last();
        }

        size_t size() const
        {
            return last() - first();
        }

    private:
        message_buffer buf_;
        const std::locale& locale_;
    };

    template<typename T>
    inline message_stream& operator<< (message_stream& out, const hex_writer<T>& mw)
    {
        mw(out);
        return out;
    }

    inline char make_readable(char c)
    {
        if (c < ' ' && c != '\t' && c != '\r' && c != '\n')
            return '?';
        else
            return c;
    }

    inline char narrow(const std::locale& loc, wchar_t c)
    {
        return make_readable(std::use_facet< std::ctype<wchar_t> >(loc).narrow(c, '?'));
    }

    inline message_stream& operator<< (message_stream& out, const wchar_t* chunk)
    {
        if (!chunk)
        {
            out << "NULL";
            return out;
        }

        for(; *chunk; ++chunk)
        {
            out << narrow(out.locale_, *chunk);
        }
        return out;
    }

    inline message_stream& operator<< (message_stream& out, const std::wstring& chunk)
    {
        for(size_t i = 0; i < chunk.size(); ++i)
        {
            out << narrow(out.locale_, chunk[i]);
        }
        return out;
    }

    inline message_stream& operator<< (message_stream& out, const char* chunk)
    {
        if (!chunk)
        {
            out << "NULL";
            return out;
        }

        for(; *chunk; ++chunk)
        {
            out << make_readable(*chunk);
        }
        return out;
    }

    inline message_stream& operator<< (message_stream& out, const std::string& chunk)
    {
        for(size_t i = 0; i < chunk.size(); ++i)
        {
            out << make_readable(chunk[i]);
        }
        return out;
    }

    template <typename T> 
    inline hex_writer<T> hexdmp(T decimal)
    {
        return hex_writer<T>(decimal);
    }
}

#endif