#include "plog/init.h"

#include "plog/monoappender_logger.h"
#include "plog/custom_log_proxy_logger.h"
#include "plog/logger_holder.h"

#include "plog/csv_formatter.h"
#include "plog/txt_formatter.h"

#include "plog/cfq_appender.h"
#include "plog/file_appender.h"
#include "plog/debug_output_appender.h"
#include "plog/ostream_appender.h"

#include "plog/error_observer.h"
#include "plog/debug_error_observer.h"

#include <memory>
#include <stdexcept>
#include <iostream>

//////////////////////////////////////////////////////////////////////////

namespace 
{
    class csv_signlefile_initializer: public plog::initializer
    {
    public:
        csv_signlefile_initializer(const wchar_t* filename, const plog::log_level level, bool rewrite_file)
            : filename_(filename)
            , level_(level)
            , rewrite_file_(rewrite_file)
        {
        }

        plog::logger* create()
        {
            std::auto_ptr<plog::file_appender> log_appender( new plog::file_appender );
            log_appender->init(filename_, rewrite_file_);
            log_appender->set_level(level_);

            return new plog::monoappender_logger( log_appender.release(), new plog::csv_formatter() );
        }

        void destroy(plog::logger* l)
        {
            if (l)
                delete l;
        }

    private:
        const wchar_t* filename_;
        const plog::log_level level_;
        bool rewrite_file_;
    };

    class txt_signlefile_initializer: public plog::initializer
    {
    public:
        txt_signlefile_initializer(const wchar_t* filename, const plog::log_level level, bool rewrite_file)
            : filename_(filename)
            , level_(level)
            , rewrite_file_(rewrite_file)
        {
        }

        plog::logger* create()
        {
            std::auto_ptr<plog::file_appender> log_appender( new plog::file_appender );
            log_appender->init(filename_, rewrite_file_);
            log_appender->set_level(level_);

            return new plog::monoappender_logger( log_appender.release(), new plog::txt_formatter() );
        }

        void destroy(plog::logger* l)
        {
            if (l)
                delete l;
        }

    private:
        const wchar_t* filename_;
        const plog::log_level level_;
        bool rewrite_file_;
    };

    class csv_cfq_initializer: public plog::initializer
    {
    public:
        csv_cfq_initializer(const wchar_t* filename, const wchar_t* extension, const plog::log_level level, size_t max_file_size, size_t max_files)
            : filename_(filename)
            , extension_(extension)
            , level_(level)
            , max_file_size_(max_file_size)
            , max_files_(max_files)
        {
        }

        plog::logger* create()
        {
            std::auto_ptr<plog::cfq_appender> log_appender(
                new plog::cfq_appender(
                    filename_, 
                    extension_, 
                    max_file_size_, 
                    max_files_,
                    level_) 
                );

            return new plog::monoappender_logger(log_appender.release(), new plog::csv_formatter());
        }

        void destroy(plog::logger* l)
        {
            if (l)
                delete l;
        }

    private:
        const wchar_t* filename_;
        const wchar_t* extension_;
        const plog::log_level level_;
        size_t max_file_size_;
        size_t max_files_;
    };

    class txt_cfq_initializer: public plog::initializer
    {
    public:
        txt_cfq_initializer(const wchar_t* filename, const wchar_t* extension, const plog::log_level level, size_t max_file_size, size_t max_files)
            : filename_(filename)
            , extension_(extension)
            , level_(level)
            , max_file_size_(max_file_size)
            , max_files_(max_files)
        {
        }

        plog::logger* create()
        {
            std::auto_ptr<plog::cfq_appender> log_appender(
                new plog::cfq_appender(
                    filename_, 
                    extension_, 
                    max_file_size_, 
                    max_files_,
                    level_) 
                );

            return new plog::monoappender_logger(log_appender.release(), new plog::txt_formatter());
        }

        void destroy(plog::logger* l)
        {
            if (l)
                delete l;
        }

    private:
        const wchar_t* filename_;
        const wchar_t* extension_;
        const plog::log_level level_;
        size_t max_file_size_;
        size_t max_files_;
    };

    class ostream_stream_initializer: public plog::initializer
    {
    public:
        ostream_stream_initializer( const plog::log_level level, 
                                    std::ostream& stream, 
                                    plog::formatter* fmtr = NULL)
            : level_(level),
              stream_(stream),
              formatter_(fmtr)
        {
            if (formatter_.get() == NULL)
            {
                formatter_ = std::auto_ptr<plog::formatter>(new plog::txt_formatter);
            }
        }

        plog::logger* create()
        {
            std::auto_ptr<plog::ostream_appender> log_appender( 
                new plog::ostream_appender(stream_, level_) );
            
            return new plog::monoappender_logger( log_appender.release(), formatter_.release() );
        }

        void destroy(plog::logger* l)
        {
            if (l)
                delete l;
        }

    private:
        const plog::log_level           level_;
        std::ostream&                   stream_;
        std::auto_ptr<plog::formatter>  formatter_;
    };

    class debug_output_initializer: public plog::initializer
    {
    public:
        debug_output_initializer(const plog::log_level level)
            : level_(level)
        {
        }

        plog::logger* create()
        {
            std::auto_ptr<plog::debug_output_appender> log_appender( new plog::debug_output_appender(level_) );
            
            return new plog::monoappender_logger( log_appender.release(), new plog::txt_formatter() );
        }

        void destroy(plog::logger* l)
        {
            if (l)
                delete l;
        }

    private:
        const plog::log_level level_;
    };

    class custom_log_initializer: public plog::initializer
    {
    public:
        custom_log_initializer(plog::custom_log* log)
            : log_(log)
        {
        }

        plog::logger* create()
        {
            return new plog::custom_log_proxy_logger(log_);
        }

        void destroy(plog::logger* l)
        {
            if (l)
                delete l;
        }

    private:
        plog::custom_log* log_;
    };
}

//////////////////////////////////////////////////////////////////////////

void plog::init(const wchar_t* filename, const plog::log_level level, bool rewrite_file)
{
    if (detail::is_csv(filename))
    {
        init_csv(filename, level, rewrite_file);
    }
    else
    {
        init_txt(filename, level, rewrite_file);
    }
}

void plog::init(const wchar_t* filename, const log_level level, size_t max_file_size, size_t max_files)
{
    if (detail::is_csv(filename))
    {
        init_csv(filename, level, max_file_size, max_files);
    }
    else
    {
        init_txt(filename, level, max_file_size, max_files);
    }
}

void plog::init_csv(const wchar_t* filename, const plog::log_level level, bool rewrite_file)
{
    std::auto_ptr<initializer> li(new csv_signlefile_initializer(filename, level, rewrite_file));
    plog::init(li);
}

void plog::init_csv(const wchar_t* filename, const log_level level, size_t max_file_size, size_t max_files)
{
    std::wstring base, extension;
    plog::detail::parse_file_name(filename, base, extension);

    std::auto_ptr<initializer> li(new csv_cfq_initializer(base.c_str(), extension.c_str(), level, max_file_size, max_files));
}

void plog::init_txt(const wchar_t* filename, const plog::log_level level, bool rewrite_file)
{
    std::auto_ptr<initializer> li(new txt_signlefile_initializer(filename, level, rewrite_file));
    plog::init(li);
}

void plog::init_txt(const wchar_t* filename, const log_level level, size_t max_file_size, size_t max_files)
{
    std::wstring base, extension;
    plog::detail::parse_file_name(filename, base, extension);

    std::auto_ptr<initializer> li(new txt_cfq_initializer(base.c_str(), extension.c_str(), level, max_file_size, max_files));
}

void plog::init_ostream(std::ostream& stream, const plog::log_level level, plog::formatter* fmtr)
{
    std::auto_ptr<initializer> li(new ostream_stream_initializer(level, stream, fmtr));
    plog::init(li);
}

void plog::init_cerr(const plog::log_level level)
{
    plog::init_ostream(std::cerr, level);
}

void plog::init_debug(const plog::log_level level)
{
    std::auto_ptr<initializer> li(new debug_output_initializer(level));
    plog::init(li);
}

void plog::init_custom_log(custom_log* log)
{
    std::auto_ptr<initializer> li(new custom_log_initializer(log));
    plog::init(li);
}

std::auto_ptr<plog::initializer> g_li;

void plog::init(std::auto_ptr<initializer> li)
{
    if (plog::get_error_observer() == NULL)
    {
        std::auto_ptr<error_observer> observer(new debug_output_error_observer);
        plog::set_error_observer(observer.release());
    }

    if (plog::logger_holder::get_logger())
    {
        plog::get_error_observer()->on_error("Log has been already initialized");
        throw std::runtime_error("Log has been already initialized");
    }

    g_li = li;
    plog::logger_holder::set_logger(g_li->create());
}

void plog::free(bool free_resources)
{
    if(! free_resources)
    {
        plog::logger_holder::release_logger();
        g_li.release();
        plog::release_error_observer();
    }
    else
    {
        if (g_li.get())
        {
            g_li->destroy(plog::logger_holder::release_logger());
            g_li.reset();
        }
        plog::set_error_observer(NULL);
    }
}

void plog::detail::parse_file_name(const wchar_t* filename, std::wstring& base, std::wstring& extension)
{
    size_t len = wcslen(filename);
    size_t lastdot = len;
    size_t lastslash = len;

    for (int i = static_cast<int>(len); i > 0; --i)
    {
        if (lastdot == len && filename[i] == '.')
        {
            lastdot = i;
        }
        if (lastslash == len && (filename[i] == '/' || filename[i] == '\\'))
        {
            lastslash = i;
        }
        if (lastdot != len && lastslash != len)
        {
            break;
        }
    }

    if (lastdot != len && (lastdot > lastslash || lastslash == len) && lastdot != lastslash + 1)
    {
        base.assign(filename, filename + lastdot);
        extension.assign(filename + lastdot);
    }
    else
    {
        base.assign(filename);
    }
}

bool plog::detail::is_csv(const wchar_t* filename)
{
    size_t len = wcslen(filename);
    return len >= 4 && !memcmp(&filename[len-4], L".csv", 4 * sizeof(wchar_t));
}