#include <boost/test/unit_test.hpp>

#include "plog/log_message.h"

#pragma warning(disable:4309)

static const std::locale loc("");

BOOST_AUTO_TEST_SUITE(message_test)

BOOST_AUTO_TEST_CASE(stream_output)
{
    std::string expectedSeverity("INFO");
    std::string expectedText("some text");

    plog::message msg(plog::info, plog::context_info(), loc, "some text");

    BOOST_REQUIRE_EQUAL(expectedSeverity, msg.severity_name());
    BOOST_REQUIRE_EQUAL(expectedText, msg.text());
}

BOOST_AUTO_TEST_CASE(check_equal)
{
    plog::message expectedMessage(plog::info, plog::context_info(), loc, "strawberry fields 4ever");

    plog::message msg(plog::info, plog::context_info(), loc);
    msg << "strawberry fields " << 4 << "ever";

    BOOST_REQUIRE_EQUAL(expectedMessage, msg);
}

BOOST_AUTO_TEST_CASE(check_different_severity)
{
    plog::message expectedMessage(plog::info, plog::context_info(), loc, "boom");

    plog::message msg(plog::error, plog::context_info(), loc);
    msg << "boom";

    BOOST_REQUIRE_NE(expectedMessage, msg);
}

BOOST_AUTO_TEST_CASE(check_different_text)
{
    plog::message expectedMessage(plog::info, plog::context_info(), loc, "foo");

    plog::message msg(plog::info, plog::context_info(), loc);
    msg << "bar";

    BOOST_REQUIRE_NE(expectedMessage, msg);
}

BOOST_AUTO_TEST_CASE(check_wide_char_const)
{
    plog::message expectedMessage(plog::info, plog::context_info(), loc, "Hello World");

    plog::message msg(plog::info, plog::context_info(), loc);
    msg << L"Hello World";

    BOOST_REQUIRE_EQUAL(expectedMessage, msg);
}

BOOST_AUTO_TEST_CASE(check_wide_char_nonconst)
{
    plog::message expectedMessage(plog::info, plog::context_info(), loc, "Hello World");

    wchar_t* nonconst = L"Hello World";
    plog::message msg(plog::info, plog::context_info(), loc);
    msg << nonconst;

    BOOST_REQUIRE_EQUAL(expectedMessage, msg);
}

BOOST_AUTO_TEST_CASE(check_wstring)
{
    plog::message expectedMessage(plog::info, plog::context_info(), loc, "Hello World");

    plog::message msg(plog::info, plog::context_info(), loc);
    msg << std::wstring(L"Hello World");

    BOOST_REQUIRE_EQUAL(expectedMessage, msg);
}

BOOST_AUTO_TEST_CASE(check_string)
{
    plog::message expectedMessage(plog::info, plog::context_info(), loc, "Hello World");

    plog::message msg(plog::info, plog::context_info(), loc);
    msg << std::string("Hello World");

    BOOST_REQUIRE_EQUAL(expectedMessage, msg);
}

BOOST_AUTO_TEST_CASE(check_char)
{
    plog::message expectedMessage(plog::info, plog::context_info(), loc, "Hello World");

    char* nonconst = "Hello World";
    plog::message msg(plog::info, plog::context_info(), loc);
    msg << nonconst;

    BOOST_REQUIRE_EQUAL(expectedMessage, msg);
}

BOOST_AUTO_TEST_CASE(check_char_const)
{
    plog::message expectedMessage(plog::info, plog::context_info(), loc, "Hello World");

    plog::message msg(plog::info, plog::context_info(), loc);
    msg << "Hello World";

    BOOST_REQUIRE_EQUAL(expectedMessage, msg);
}

BOOST_AUTO_TEST_CASE(string_containing_0)
{
    std::string str = "hello";
    str += '\0';
    str += "this must be logged";

    plog::message msg(plog::info, plog::context_info(), loc);

    msg << str;

    BOOST_REQUIRE_EQUAL("hello?this must be logged", msg.text());
}

BOOST_AUTO_TEST_CASE(check_hex)
{
    plog::message expectedMessage(plog::info, plog::context_info(), loc, "DEADBEEF");

    plog::message msg(plog::info, plog::context_info(), loc);
    msg << plog::hexdmp(0xDEADBEEF);

    BOOST_CHECK_EQUAL(expectedMessage, msg);
}

BOOST_AUTO_TEST_CASE(check_hex_int)
{
    plog::message expectedMessage(plog::info, plog::context_info(), loc, "00001234");

    plog::message msg(plog::info, plog::context_info(), loc);
    msg << plog::hexdmp(0x1234);

    BOOST_CHECK_EQUAL(expectedMessage, msg);
}

BOOST_AUTO_TEST_CASE(check_hex_short)
{
    plog::message expectedMessage(plog::info, plog::context_info(), loc, "0123");

    plog::message msg(plog::info, plog::context_info(), loc);
    msg << plog::hexdmp<short>(0x123);

    BOOST_CHECK_EQUAL(expectedMessage, msg);
}

BOOST_AUTO_TEST_CASE(check_hex_char)
{
    plog::message expectedMessage(plog::info, plog::context_info(), loc, "AB");

    plog::message msg(plog::info, plog::context_info(), loc);
    msg << plog::hexdmp<char>(0xAB);

    BOOST_CHECK_EQUAL(expectedMessage, msg);
}

BOOST_AUTO_TEST_SUITE_END()
