#include <boost/test/unit_test.hpp>

#include "plog/txt_formatter.h"


namespace
{
    const char short_msg[] = "short message is short";

    const char long_msg[] = 
        "long message is so \
looooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo\
oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo\
oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo\
oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo\
oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo\
oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo\
oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo\
oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooong";
}

    const char multiline_msg[] = 
"Hard kitty, cold kitty,\n\
Little ball of steel.\n\
Evil kitty, angry kitty,\n\
Kill, kill, kill...";

static const size_t BUF_SIZE = 512;
static char buffer[BUF_SIZE];
static const std::locale loc("");

BOOST_AUTO_TEST_SUITE(txt_formatter)

BOOST_AUTO_TEST_CASE(format_short_message)
{
    SYSTEMTIME time = {1986, 9, 1, 1, 12, 0, 0, 0};
    plog::message msg(
        plog::info, 
        plog::context_info(__FILE__, __FUNCTION__, 777, 1234, NULL, &time), 
        loc,
        short_msg);

    plog::txt_formatter formatter;
    size_t length = formatter.format(msg, buffer, BUF_SIZE);

    std::stringstream expected;
    expected << "1986.09.01\t" << "12:00:00:000\t[   INFO]\t" << "1234\t" << __FUNCTION__ << "\t" << short_msg << "\r\n";
    size_t expectedLength = expected.str().length();
    
    BOOST_CHECK_EQUAL(expectedLength, length);
    BOOST_CHECK_EQUAL(buffer[length - 2], '\r');
    BOOST_CHECK_EQUAL(buffer[length - 1], '\n');
    BOOST_CHECK_EQUAL(expected.str(), buffer);
}

BOOST_AUTO_TEST_CASE(format_long_message)
{
    SYSTEMTIME time = {1986, 9, 1, 1, 12, 0, 0, 0};
    plog::message msg(
        plog::info, 
        plog::context_info(__FILE__, __FUNCTION__, 777, 1234, NULL, &time), 
        loc,
        long_msg);

    plog::txt_formatter formatter;
    size_t length = formatter.format(msg, buffer, BUF_SIZE);

    std::stringstream expected;
    expected << "1986.09.01\t" << "12:00:00:000\t[   INFO]\t" << "1234\t" << __FUNCTION__ << "\t" << long_msg << "\r\n";
    size_t expectedLength = 511;

    BOOST_CHECK_EQUAL(expectedLength, length);
    BOOST_CHECK_EQUAL(expected.str().substr(0, expectedLength - 2), std::string(buffer).substr(0, expectedLength - 2));
    BOOST_CHECK_EQUAL(buffer[length - 2], '\r');
    BOOST_CHECK_EQUAL(buffer[length - 1], '\n');
    BOOST_CHECK_EQUAL(expectedLength, length);
}

BOOST_AUTO_TEST_CASE(multiline_message)
{
    SYSTEMTIME time = {1986, 9, 1, 1, 12, 0, 0, 0};
    plog::message msg(
        plog::info, 
        plog::context_info(__FILE__, __FUNCTION__, 777, 1234, NULL, &time), 
        loc,
        multiline_msg);

    plog::txt_formatter formatter;
    size_t length = formatter.format(msg, buffer, BUF_SIZE);

    std::stringstream expected;
    expected << "1986.09.01\t" << "12:00:00:000\t[   INFO]\t" << "1234\t" << __FUNCTION__ << "\t" << "Hard kitty, cold kitty," << "\r\n";
    expected << "1986.09.01\t" << "12:00:00:000\t[   INFO]\t" << "1234\t" << __FUNCTION__ << "\t" << "Little ball of steel." << "\r\n";
    expected << "1986.09.01\t" << "12:00:00:000\t[   INFO]\t" << "1234\t" << __FUNCTION__ << "\t" << "Evil kitty, angry kitty," << "\r\n";
    expected << "1986.09.01\t" << "12:00:00:000\t[   INFO]\t" << "1234\t" << __FUNCTION__ << "\t" << "Kill, kill, kill..." << "\r\n";
    size_t expectedLength = expected.str().length();

    BOOST_CHECK_EQUAL(expectedLength, strlen(buffer));
    BOOST_CHECK_EQUAL(buffer[length - 2], '\r');
    BOOST_CHECK_EQUAL(buffer[length - 1], '\n');
    BOOST_CHECK_EQUAL(expected.str(), buffer);

}

BOOST_AUTO_TEST_CASE(cr_lf_crash)
{
    SYSTEMTIME time = {1986, 9, 1, 1, 12, 0, 0, 0};
    plog::message msg(
        plog::info, 
        plog::context_info(__FILE__, __FUNCTION__, 777, 1234, NULL, &time), 
        loc,
        "Crash\r\n");

    plog::txt_formatter formatter;
    formatter.format(msg, buffer, BUF_SIZE);
}
BOOST_AUTO_TEST_CASE(cr_lf_correct)
{
    SYSTEMTIME time = {1986, 9, 1, 1, 12, 0, 0, 0};
    plog::message msg(
        plog::info, 
        plog::context_info(__FILE__, __FUNCTION__, 777, 1234, NULL, &time), 
        loc,
        "string1\r\nstring2");

    plog::txt_formatter formatter;
    formatter.format(msg, buffer, BUF_SIZE);
    std::stringstream expected;
    expected << "1986.09.01\t" << "12:00:00:000\t[   INFO]\t" << "1234\t" << __FUNCTION__ << "\t" << "string1" << "\r\n";
    expected << "1986.09.01\t" << "12:00:00:000\t[   INFO]\t" << "1234\t" << __FUNCTION__ << "\t" << "string2" << "\r\n";
    BOOST_CHECK_EQUAL(expected.str(), buffer);
}

BOOST_AUTO_TEST_CASE(format_trace)
{
    SYSTEMTIME time = {1986, 9, 1, 1, 12, 0, 0, 0};
    plog::message msg(
        plog::trace, 
        plog::context_info(__FILE__, __FUNCTION__, 777, 1234, NULL, &time), 
        loc,
        "");

    plog::txt_formatter formatter;
    size_t length = formatter.format(msg, buffer, BUF_SIZE);

    std::stringstream expected;
    expected << "1986.09.01\t" << "12:00:00:000\t[  TRACE]\t" << "1234\t" << __FUNCTION__ "\t\r\n";
    size_t expectedLength = expected.str().length();
    
    BOOST_CHECK_EQUAL(expectedLength, length);
    BOOST_CHECK_EQUAL(buffer[length - 2], '\r');
    BOOST_CHECK_EQUAL(buffer[length - 1], '\n');
    BOOST_CHECK_EQUAL(expected.str(), buffer);
}

BOOST_AUTO_TEST_SUITE_END()
